/*
 *	Module: File Node
 *	Function: Create a file node when a file object creation occurs, then tell
 *		file if it is protected. Free the file node while closing the relative file
 *		object. All strings are wide charicters.
 *	Author: Huang Guan
 *	Email: gdxxhg at gmail.com
 *	Date: 23/12/2007 
 */

#include <stdio.h>
#include <string.h>
#include <wchar.h>
#include <stdlib.h>
#include <windows.h>
#include "../sys/file_node.h"

#define dprint printf
#define kmem_get malloc
#define kmem_free free

struct file_node{
	//link table
	short next;	//next id of the file_node.
	short pre;		//previous id of the file_node.
	short parent;	//parent id of the file_node.
	short child;	//child id of the file_node.
	//internal information
	wchar_t file_name[MAX_FILENAME_LEN];
	unsigned char prot_type;	//indicate the type of protection. ( PT_RDONLY )
	unsigned char prot_flag;   //means only protect executable files
	unsigned char node_type;   //type of node. NT_NORMAL or NT_CUSTOMED
	short  count;	//how many users using this file, the normal node will be removed when count==0.
};

//static vars
static struct file_node *file_node = NULL;
static int nr_nodes;	//number of the using file nodes
static struct file_node *fn_root;
static int fs_status;
//static functions
static int remove_file_node( short target );

//do the initialization
int file_node_init()
{
	//allocate memories...
	int size_alloc;
	size_alloc = sizeof( struct file_node ) * MAX_FILENODE;
	//dprint("alloc memory: %dB\n", size_alloc );
	file_node = (struct file_node*) kmem_get( size_alloc );
	memset( file_node, 0, size_alloc );
	if( file_node == NULL )
	{
		dprint("No enough memory.\n");
		return -1;
	}
	fn_root = &file_node[1]; //note: we use element 1 as the root file node.
	fn_root -> next = fn_root ->pre = fn_root ->parent = fn_root ->child = 0;
	fn_root -> prot_type = fn_root -> prot_flag = 0;
	fn_root -> count = 1;
	wcsncpy( fn_root -> file_name, L"root", MAX_FILENAME_LEN );
	nr_nodes = 1;
	//ok
	//enter protection mode.
	prot_status( PS_PROTECT );
	return 0;
}

//get a free node, return < 0 means failed.
static short get_free_file_node()
{
	short i;
	if( file_node == NULL ) return -1;
	for( i=2; i<MAX_FILENODE; i++ )
	{
		if( file_node[i].count == 0 )	//will this cause race condition???
		{
			file_node[i].count = 1;
			nr_nodes++;
			return i;
		}
		#ifndef	RELEASE
		if( file_node[i].count<0 )
		{
			dprint("impossible: file_node[%d].count=%d\n", i, file_node[i].count );
		}
		#endif
	}
	dprint("no enough file_node.\n");
	return -1;
}

//return a file node by name in a detemined directory node.
static short get_file_node_by_name( short dir, wchar_t* name )
{
	short a;
	if( dir<0 ) 
	   return -1;
	for( a = file_node[dir].child; a; a = file_node[a].next ){
		if( _wcsnicmp( file_node[a].file_name, name, MAX_FILENAME_LEN ) == 0 )
		{
			return a;
		}
	}
	return -1;	//not found.
}

//build a new file node within  a parent node.
static short build_file_node( short dir, wchar_t* name )
{
	short target;
	target = get_free_file_node();
	if( target==-1 )
		return  -1;
	//make link
	file_node[target].pre = 0;
	file_node[target].parent = dir;
	file_node[target].next = file_node[dir].child;
	file_node[ file_node[target].next ].pre = target;
	file_node[dir].child = target;
	//initialize
	wcsncpy( file_node[target].file_name, name, MAX_FILENAME_LEN );
	file_node[target].node_type = NT_NORMAL;
	file_node[target].prot_type = PT_NONE;
	file_node[target].prot_flag = 0;
	nr_nodes ++;
	return target;
}

//remove protection point
int remove_prot_node( short target )
{
	if( target < 0 )
		return -1;
	file_node[target].prot_flag = 0;
	file_node[target].prot_type = 0;
	file_node[target].node_type = NT_NORMAL;
	remove_file_node( target );
	return 0;
}

//remove the file node by node id.
static int remove_file_node( short target )
{
	short pr;
	if( target<=0 )
	{
		dprint("Warning: removing file node %d\n", target );
		return -1;
	}
	if( file_node[target].count <= 0 )
	{
		dprint("%ls, Warning: file_node count is : %d\n", 
			file_node[target].file_name, file_node[target].count );
		return -1;
	}
	file_node[target].count --;
	pr = file_node[target].parent;
	if( file_node[target].count == 0 ){
		//make link
		//dprint("delete node: %ls\n", file_node[target].file_name );
		if( file_node[target].child>0 )
		{	
			dprint("%ls, Warning: file_node child is %d\n", 
				file_node[target].file_name, file_node[target].child );
		}
		file_node[ file_node[target].next ].pre = file_node[target].pre;
		if( file_node[target].pre == 0 ) //direct child
		{
			file_node[pr].child = file_node[target].next;
		}else{
			file_node[ file_node[target].pre ].next = file_node[target].next;
		}
		nr_nodes--;
	}
	if( pr )
		remove_file_node( pr );
	return 0;
}

//separate a path string and make nodes.
static short create_path( wchar_t *path )
{
	//  C:\boot\s\31.txt ==> C:, boot, s, 31.txt
	//  D:\Windows ==> D:, Windows
	wchar_t* tmp, *token;
	short cur, target;
	wchar_t file_name[MAX_FILENAME_LEN];
	if( path[0] == (wchar_t)0 ) //empty path
		return -1;
	tmp = path;
	//at first it begins from root.
	cur = 1;
	file_node[ cur ].count ++;
	//dprint("Path: %ls\n", path );
	for(;tmp;){
		token = wcschr( tmp, '\\' );
		if( token==NULL ){	//not found\end
			wcsncpy( file_name, tmp, MAX_FILENAME_LEN );
			tmp=NULL;
		}else{
			int len;
			len = (int)(token-tmp);
			wcsncpy( file_name, tmp, len );
			file_name[ len ] = (wchar_t)0;
			tmp = token+1;
		}
		//dprint("file_name: %ls\n", file_name );
		target = get_file_node_by_name( cur, file_name );  //search to find if this node has been added.
		if( target<0 ){ //not exists, now add it.
			target = build_file_node( cur, file_name );
		}else{
			file_node[target].count++;	//inc reference
		}
		cur = target;
	}
	//over, file_name is the file name, cur is its node
	return cur;
}


//add a protected node
int add_prot_node( wchar_t *path, unsigned char type, unsigned char prot_flag )
{
	short i;
	if( file_node == NULL )
	{
		return -1;
	}
	i = create_path( path );
	if( i>0 ){
		file_node[i].prot_type = type;
		file_node[i].prot_flag = prot_flag;
	}
	file_node[i].node_type = NT_CUSTOMED;
	return i;
}

//this will be called in the end.
int file_node_cleanup()
{
	//dprint("file_node_cleanup();\n");
	display_file_node();
	kmem_free( file_node );
	file_node = NULL;
	nr_nodes = 0;
	return 0;
}


//display all nodes
static void __display_file_node( int i, int level )
{
	int j;
	char pad[10];
	for( ; i; i=file_node[i].next )
	{
		for( j=0; j<level && j<10; j++ ) pad[j]=' ';
		pad[j] = '\0';
		dprint("%s|--[%d:%ls:%d:%d](%d)\n", pad, i, file_node[i].file_name, 
			file_node[i].prot_type, file_node[i].prot_flag, file_node[i].count );
		__display_file_node( file_node[i].child, level+1 );
	}
}

static void _deep_search( FILE* fp, short i, char* path, int pos )
{
	short t;
	char name[MAX_FILENAME_LEN];
	for( t = file_node[i].child; t; t = file_node[t].next )
	{
		int len;
		path[pos]=0;
		len = WideCharToMultiByte( CP_ACP, WC_COMPOSITECHECK, 
			file_node[t].file_name, -1, name, sizeof(name), NULL, NULL );
		name[len] = 0;
		sprintf( path, "%s\\%s", path, name );
		if( file_node[t].node_type == NT_CUSTOMED )
		{
			fprintf( fp, "%s\t%d\t%d\n", path+1, file_node[t].prot_type, 
				file_node[t].prot_flag );
		}
		_deep_search( fp, t, path, strlen(path) );
	}
}

//write out file nodes, make them into paths
void out_file_node( char* file )
{
	FILE *fp = fopen( file, "w" );
	char path[2048];
	path[0]=0;
	if( fp != NULL )
	{
		_deep_search( fp, 1, path, 0 );
		fclose( fp );
		return;
	}
	printf("Error: can't open file %s\n", file );
} 

void display_file_node()
{
	if( file_node==NULL )
		return;
	__display_file_node( 1, 0 );
}

//set protection status
int prot_status( int status )
{
	switch( status )
	{
		case PS_IGNORE:
			fs_status = status;
			//dprint("## Protection stopped!\n");
			break;
		case PS_PROTECT: 
			fs_status = status;
			//dprint("## Protection started!\n");
			break;
			//dprint("## Unknown status!\n");
	}
	return fs_status;
}

//load file nodes
int set_file_nodes( char* buffer, int len )
{
	int cb;
	cb = sizeof( struct file_node ) * MAX_FILENODE;
	if ( len <= cb )
	{
		if( file_node==NULL )
			return 0;
		memcpy( file_node, buffer, len );
		nr_nodes = len/sizeof( struct file_node );
		printf("nr_nodes= %d\n", nr_nodes);
		return cb;
	}
	return 0;
}

//get file node buffer
int get_file_nodes( char* buffer, int len )
{
	int cb;
	cb = sizeof( struct file_node ) * nr_nodes;
	if ( len >= cb )
	{
		if( file_node==NULL )
			return 0;
		memcpy( buffer, file_node, cb );
		return cb;
	}
	return 0;
}

